# -*- coding: utf-8 -*-

# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hvac
import mock
import os
import uuid

from oslotest import base
from testtools import testcase


TEST_POLICY = '''
path "{backend}/*" {{
  capabilities = ["create", "read", "update", "delete", "list"]
}}
'''


class VaultlockerFuncBaseTestCase(base.BaseTestCase):

    """Test case base class for all functional tests."""

    def setUp(self):
        super(VaultlockerFuncBaseTestCase, self).setUp()
        self.vault_client = None

        self.vault_addr = os.environ.get('PIFPAF_VAULT_ADDR')
        self.root_token = os.environ.get('PIFPAF_ROOT_TOKEN')

        self.test_uuid = str(uuid.uuid4())
        self.vault_backend = 'vaultlocker-test-{}'.format(self.test_uuid)
        self.vault_policy = 'vaultlocker-policy-{}'.format(self.test_uuid)
        self.vault_approle = 'vaultlocker-approle-{}'.format(self.test_uuid)

        if not self.vault_addr or not self.root_token:
            raise testcase.TestSkipped('Vault not running')

        self.vault_client = hvac.Client(url=self.vault_addr,
                                        token=self.root_token)

        self.vault_client.enable_secret_backend(
            backend_type='kv',
            description='vault test backend',
            mount_point=self.vault_backend
        )

        try:
            self.vault_client.enable_auth_backend('approle')
        except hvac.exceptions.InvalidRequest:
            pass

        self.vault_client.set_policy(
            name=self.vault_policy,
            rules=TEST_POLICY.format(backend=self.vault_backend)
        )

        self.vault_client.create_role(
            self.vault_approle,
            token_ttl='60s',
            token_max_ttl='60s',
            policies=[self.vault_policy],
            bind_secret_id='true',
            bound_cidr_list='127.0.0.1/32')
        self.approle_uuid = self.vault_client.get_role_id(self.vault_approle)
        self.secret_id = self.vault_client.write(
            'auth/approle/role/{}/secret-id'.format(self.vault_approle)
        )['data']['secret_id']

        self.test_config = {
            'vault': {
                'url': self.vault_addr,
                'approle': self.approle_uuid,
                'secret_id': self.secret_id,
                'backend': self.vault_backend,
            }
        }
        self.config = mock.MagicMock()
        self.config.get.side_effect = \
            lambda s, k: self.test_config.get(s).get(k)

    def tearDown(self):
        super(VaultlockerFuncBaseTestCase, self).tearDown()
        if self.vault_client:
            self.vault_client.disable_secret_backend(self.vault_backend)
            self.vault_client.delete_policy(self.vault_policy)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_systemd
----------------------------------

Tests for `systemd` module.
"""

import mock

from vaultlocker import systemd
from vaultlocker.tests.unit import base


class TestSystemD(base.TestCase):

    @mock.patch.object(systemd, 'subprocess')
    def test_enable(self, _subprocess):
        systemd.enable('my-service.service')
        _subprocess.check_call.assert_called_once_with(
            ['systemctl', 'enable', 'my-service.service']
        )



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_dmcrypt
----------------------------------

Tests for `dmcrypt` module.
"""

import base64
import mock

from vaultlocker import dmcrypt
from vaultlocker.tests.unit import base


class TestDMCrypt(base.TestCase):

    @mock.patch.object(dmcrypt, 'subprocess')
    def test_luks_format(self, _subprocess):
        dmcrypt.luks_format('mykey', '/dev/sdb', 'test-uuid')
        _subprocess.check_output.assert_called_once_with(
            ['cryptsetup',
             '--batch-mode',
             '--uuid', 'test-uuid',
             '--key-file', '-',
             'luksFormat', '/dev/sdb'],
            input='mykey'.encode('UTF-8')
        )

    @mock.patch.object(dmcrypt, 'subprocess')
    def test_luks_open(self, _subprocess):
        dmcrypt.luks_open('mykey', 'test-uuid')
        _subprocess.check_output.assert_called_once_with(
            ['cryptsetup',
             '--batch-mode',
             '--key-file', '-',
             'open', 'UUID=test-uuid', 'crypt-test-uuid',
             '--type', 'luks'],
            input='mykey'.encode('UTF-8')
        )

    @mock.patch.object(dmcrypt, 'os')
    def test_generate_key(self, _os):
        _key = b'randomdatastringfromentropy'
        _os.urandom.return_value = _key
        self.assertEqual(dmcrypt.generate_key(),
                         base64.b64encode(_key).decode('UTF-8'))
        _os.urandom.assert_called_with(dmcrypt.KEY_SIZE / 8)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import hvac
import logging
import os
import socket
import tenacity
import uuid

from six.moves import configparser

from vaultlocker import dmcrypt
from vaultlocker import systemd

logger = logging.getLogger(__name__)

RUN_VAULTLOCKER = '/run/vaultlocker'
CONF_FILE = '/etc/vaultlocker/vaultlocker.conf'


def _vault_client(config):
    """Helper wrapper to create Vault Client

    :param: config: configparser object of vaultlocker config
    :returns: hvac.Client. configured Vault Client object
    """
    client = hvac.Client(url=config.get('vault', 'url'))
    client.auth_approle(config.get('vault', 'approle'),
                        secret_id=config.get('vault', 'secret_id'))
    return client


def _get_vault_path(device_uuid, config):
    """Generate full vault path for a given block device UUID

    :param: device_uuid: String of the device UUID
    :param: config: configparser object of vaultlocker config
    :returns: str: Path to vault resource for device
    """
    return '{}/{}/{}'.format(config.get('vault', 'backend'),
                             socket.gethostname(),
                             device_uuid)


def _encrypt_block_device(args, client, config):
    """Encrypt and open a block device

    Stores the dm-crypt key direct in vault

    :param: args: argparser generated cli arguments
    :param: client: hvac.Client for Vault access
    :param: config: configparser object of vaultlocker config
    """
    block_device = args.block_device[0]
    key = dmcrypt.generate_key()
    block_uuid = str(uuid.uuid4()) if not args.uuid else args.uuid
    vault_path = _get_vault_path(block_uuid, config)

    dmcrypt.luks_format(key, block_device, block_uuid)
    # Ensure sym link for new encrypted device is created
    # LP Bug #1780332
    dmcrypt.udevadm_settle(block_uuid)

    # NOTE: store and validate key
    client.write(vault_path,
                 dmcrypt_key=key)
    stored_data = client.read(vault_path)
    assert key == stored_data['data']['dmcrypt_key']

    dmcrypt.luks_open(key, block_uuid)

    systemd.enable('vaultlocker-decrypt@{}.service'.format(block_uuid))


def _decrypt_block_device(args, client, config):
    """Open a LUKS/dm-crypt encrypted block device

    The devices dm-crypt key is retrieved from Vault

    :param: args: argparser generated cli arguments
    :param: client: hvac.Client for Vault access
    :param: config: configparser object of vaultlocker config
    """
    block_uuid = args.uuid[0]
    vault_path = _get_vault_path(block_uuid, config)

    stored_data = client.read(vault_path)
    if stored_data is None:
        raise ValueError('Unable to locate key for {}'.format(block_uuid))
    key = stored_data['data']['dmcrypt_key']

    dmcrypt.luks_open(key, block_uuid)


def _do_it_with_persistence(func, args, config):
    """Exec func with retries based on provided cli flags

    :param: func: function to attempt to execute
    :param: args: argparser generated cli arguments
    :param: config: configparser object of vaultlocker config
    """
    @tenacity.retry(
        wait=tenacity.wait_fixed(1),
        reraise=True,
        stop=(
            tenacity.stop_after_delay(args.retry) if args.retry > 0
            else tenacity.stop_after_attempt(1)
            ),
        retry=(
            tenacity.retry_if_exception(hvac.exceptions.VaultNotInitialized) |
            tenacity.retry_if_exception(hvac.exceptions.VaultDown)
            )
        )
    def _do_it():
        client = _vault_client(config)
        func(args, client, config)
    _do_it()


def encrypt(args, config):
    """Encrypt and open handler

    :param: args: argparser generated cli arguments
    :param: config: configparser object of vaultlocker config
    """
    _do_it_with_persistence(_encrypt_block_device, args, config)


def decrypt(args, config):
    """Decrypt and open handler

    :param: args: argparser generated cli arguments
    :param: config: configparser object of vaultlocker config
    """
    _do_it_with_persistence(_decrypt_block_device, args, config)


def get_config():
    """Read vaultlocker configuration from config file

    :returns: configparser. Parsed configuration options
    """
    config = configparser.ConfigParser()
    if os.path.exists(CONF_FILE):
        config.read(CONF_FILE)
    return config


def main():
    parser = argparse.ArgumentParser('vaultlocker')
    parser.set_defaults(prog=parser.prog)
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="sub-command help",
    )
    parser.add_argument(
        '--retry',
        default=-1,
        type=int,
        help="Time in seconds to continue retrying to connect to Vault"
    )

    encrypt_parser = subparsers.add_parser(
        'encrypt',
        help='Encrypt a block device and store its key in Vault'
    )
    encrypt_parser.add_argument('--uuid',
                                dest="uuid",
                                help="UUID to use to reference encryption key")
    encrypt_parser.add_argument('block_device',
                                metavar='BLOCK_DEVICE', nargs=1,
                                help="Full path to block device to encrypt")
    encrypt_parser.set_defaults(func=encrypt)

    decrypt_parser = subparsers.add_parser(
        'decrypt',
        help='Decrypt a block device retrieving its key from Vault'
    )
    decrypt_parser.add_argument('uuid',
                                metavar='uuid', nargs=1,
                                help='UUID of block device to decrypt')
    decrypt_parser.set_defaults(func=decrypt)

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    try:
        args.func(args, get_config())
    except Exception as e:
        raise SystemExit(
            '{prog}: {msg}'.format(
                prog=args.prog,
                msg=e,
            )
        )



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import logging
import os
import subprocess

logger = logging.getLogger(__name__)


KEY_SIZE = 4096


def generate_key():
    """Generate a 4096 bit random key for use with dm-crypt

    :returns: str.  Base64 encoded 4096 bit key
    """
    data = os.urandom(int(KEY_SIZE / 8))
    key = base64.b64encode(data).decode('utf-8')
    return key


def luks_format(key, device, uuid):
    """LUKS format a block device

    Format a block device using dm-crypt/LUKS with the
    provided key and uuid

    :param: key: string containing the encryption key to use.
    :param: device: full path to block device to use.
    :param: uuid: uuid to use for encrypted block device.
    """
    logger.info('LUKS formatting {} using UUID:{}'.format(device, uuid))
    command = [
        'cryptsetup',
        '--batch-mode',
        '--uuid',
        uuid,
        '--key-file',
        '-',
        'luksFormat',
        device,
    ]
    subprocess.check_output(command,
                            input=key.encode('UTF-8'))


def luks_open(key, uuid):
    """LUKS open a block device by UUID

    Open a block device using dm-crypt/LUKS with the
    provided key and uuid

    :param: key: string containing the encryption key to use.
    :param: uuid: uuid to use for encrypted block device.
    :returns: str. dm-crypt mapping
    """
    logger.info('LUKS opening {}'.format(uuid))
    handle = 'crypt-{}'.format(uuid)
    command = [
        'cryptsetup',
        '--batch-mode',
        '--key-file',
        '-',
        'open',
        'UUID={}'.format(uuid),
        handle,
        '--type',
        'luks',
    ]
    subprocess.check_output(command,
                            input=key.encode('UTF-8'))
    return handle


def udevadm_settle(uuid):
    """udevadm settle the newly created encrypted device

    Ensure udev has created symlink for newly created encyprted device.
    See LP Bug #1780332

    :param: uuid: uuid to use for encrypted block device.
    """
    logger.info('udevadm settle /dev/disk/by-uuid/{}'.format(uuid))
    command = [
        'udevadm',
        'settle',
        '--exit-if-exists=/dev/disk/by-uuid/{}'.format(uuid),
    ]
    subprocess.check_output(command)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_vaultlocker
----------------------------------

Tests for `vaultlocker` module.
"""

import mock

from vaultlocker import shell
from vaultlocker.tests.unit import base


class TestVaultlocker(base.TestCase):

    _test_config = {
        'url': 'https://vaultlocker.test.com',
        'approle': '85e4c349-7547-4ad5-9172-d82a45d87b3e',
        'secret_id': '9428ad25-7b4a-442f-8f20-f23be0575146',
        'backend': 'vaultlocker-test',
    }

    def __init__(self, *args, **kwds):
        super(TestVaultlocker, self).__init__(*args, **kwds)
        self.config = mock.MagicMock()
        self.config.get.side_effect = lambda _, k: self._test_config.get(k)

    @mock.patch.object(shell, 'systemd')
    @mock.patch.object(shell, 'dmcrypt')
    @mock.patch.object(shell, '_get_vault_path')
    def test_encrypt(self, _get_vault_path, _dmcrypt, _systemd):
        _get_vault_path.return_value = 'backend/host/uuid'
        _dmcrypt.generate_key.return_value = 'testkey'

        args = mock.MagicMock()
        args.uuid = 'passed-UUID'
        args.block_device = ['/dev/sdb']

        client = mock.MagicMock()
        client.read.return_value = {
            'data': {
                'dmcrypt_key': 'testkey'
            }
        }

        shell._encrypt_block_device(args, client, self.config)

        _dmcrypt.luks_format.assert_called_with(
            'testkey', '/dev/sdb', 'passed-UUID'
        )
        _dmcrypt.luks_open.assert_called_with(
            'testkey', 'passed-UUID'
        )
        _systemd.enable.assert_called_with(
            'vaultlocker-decrypt@passed-UUID.service'
        )

    @mock.patch.object(shell, 'systemd')
    @mock.patch.object(shell, 'dmcrypt')
    @mock.patch.object(shell, '_get_vault_path')
    def test_encrypt_vault_failure(self, _get_vault_path,
                                   _dmcrypt, _systemd):
        _get_vault_path.return_value = 'backend/host/uuid'
        _dmcrypt.generate_key.return_value = 'testkey'

        args = mock.MagicMock()
        args.uuid = 'passed-UUID'
        args.block_device = ['/dev/sdb']

        client = mock.MagicMock()
        client.read.return_value = {
            'data': {
                'dmcrypt_key': 'brokendata'
            }
        }

        self.assertRaises(
            AssertionError,
            shell._encrypt_block_device,
            args, client, self.config
        )

    @mock.patch.object(shell, 'dmcrypt')
    @mock.patch.object(shell, '_get_vault_path')
    def test_decrypt(self, _get_vault_path, _dmcrypt):
        _get_vault_path.return_value = 'backend/host/uuid'

        args = mock.MagicMock()
        args.uuid = ['passed-UUID']

        client = mock.MagicMock()
        client.read.return_value = {
            'data': {
                'dmcrypt_key': 'testkey'
            }
        }

        shell._decrypt_block_device(args, client, self.config)

        _dmcrypt.luks_open.assert_called_with(
            'testkey', 'passed-UUID'
        )

    @mock.patch.object(shell, 'socket')
    def test_get_vault_path(self, _socket):
        _socket.gethostname.return_value = 'myhost'
        self.assertEqual(shell._get_vault_path('my-UUID', self.config),
                         'vaultlocker-test/myhost/my-UUID')



# -*- coding: utf-8 -*-

# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from vaultlocker import shell
from vaultlocker.tests.functional import base


@mock.patch.object(shell, 'systemd')
@mock.patch.object(shell.dmcrypt, 'luks_format')
@mock.patch.object(shell.dmcrypt, 'luks_open')
class KeyStorageTestCase(base.VaultlockerFuncBaseTestCase):

    """Test storage and retrieval of dm-crypt keys from vault"""

    def test_encrypt(self, _luks_open, _luks_format, _systemd):
        """Test encrypt function stores correct data in vault"""
        args = mock.MagicMock()
        args.uuid = 'passed-UUID'
        args.block_device = ['/dev/sdb']
        args.retry = -1

        shell.encrypt(args, self.config)
        _luks_format.assert_called_once_with(mock.ANY,
                                             '/dev/sdb',
                                             'passed-UUID')
        _luks_open.assert_called_once_with(mock.ANY,
                                           'passed-UUID')
        _systemd.enable.assert_called_once_with(
            'vaultlocker-decrypt@passed-UUID.service'
        )

        stored_data = self.vault_client.read(
            shell._get_vault_path('passed-UUID',
                                  self.config)
        )
        self.assertIsNotNone(stored_data,
                             'Key data missing from vault')
        self.assertTrue('dmcrypt_key' in stored_data['data'],
                        'dm-crypt key data is missing')

    def test_decrypt(self, _luks_open, _luks_format, _systemd):
        """Test decrypt function retrieves correct key from vault"""
        args = mock.MagicMock()
        args.uuid = ['passed-UUID']
        args.retry = -1

        self.vault_client.write(shell._get_vault_path('passed-UUID',
                                                      self.config),
                                dmcrypt_key='testkey')

        shell.decrypt(args, self.config)
        _luks_format.assert_not_called()
        _systemd.enable.assert_not_called()
        _luks_open.assert_called_once_with('testkey',
                                           'passed-UUID')

    def test_decrypt_missing_key(self, _luks_open, _luks_format, _systemd):
        """Test decrypt function errors if a key is missing from vault"""
        args = mock.MagicMock()
        args.uuid = ['passed-UUID']
        args.retry = -1

        self.assertRaises(ValueError,
                          shell.decrypt,
                          args, self.config)
        _luks_format.assert_not_called()
        _systemd.enable.assert_not_called()
        _luks_open.assert_not_called()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import subprocess

logger = logging.getLogger(__name__)


def enable(service_name):
    """Enable a systemd unit

    :param: service_name: Name of the service to enable.
    """
    logging.info('Enabling systemd unit for {}'.format(service_name))
    cmd = ['systemctl', 'enable', service_name]
    subprocess.check_call(cmd)
